//
//	@file
//	robotout - a max object (Arduino UNO board writer)
//
// written by Orly Natan, Creative Robotics Lab
// University Of New South Wales (UNSW). */


// required Max core headers
#include "ext.h"							// standard Max include, always required
#include "ext_obex.h"						// required for new style Max object

#include "wrapper.h"

////////////////////////// object struct
typedef struct robotout
{
    // the Max object itself (must be first)
    t_object                    ob;
    
    // robotout attributes:
    t_symbol                    *ip_addr;
    Float64                     inlets[6];
    
    // robotout state
    bool                        is_ready;
    
} t_robotout;

///////////////////////// function prototypes
//// standard set
void *robotout_new(t_symbol *s, long argc, t_atom *argv);
void robotout_free(t_robotout *x);
void robotout_assist(t_robotout *x, void *b, long m, long a, char *s);
void robotout_bang(t_robotout *x);
int  robotout_open(t_robotout *x);
int  robotout_close(t_robotout *x);
void robotout_break(t_robotout *x);
void robotout_ft1(t_robotout *x, Float64 f);
void robotout_ft2(t_robotout *x, Float64 f);
void robotout_ft3(t_robotout *x, Float64 f);
void robotout_ft4(t_robotout *x, Float64 f);
void robotout_ft5(t_robotout *x, Float64 f);
void robotout_ft6(t_robotout *x, Float64 f);

int robotout_parseArgs(t_robotout *x, long argc, t_atom *argv);

//////////////////////// global class pointer variable
void *robotout_class;


void ext_main(void *r)
{
    t_class *c;
    
    c = class_new("cr.robotout", (method)robotout_new, (method)robotout_free,
                  (long)sizeof(t_robotout),
                  0L /* leave NULL!! */, A_GIMME, 0);
    
    class_addmethod(c, (method)robotout_assist,      "assist",  A_CANT,     0);
    class_addmethod(c, (method)robotout_open,        "open",                0);
    class_addmethod(c, (method)robotout_close,       "close",               0);
    class_addmethod(c, (method)robotout_bang,        "bang",                0);
    class_addmethod(c, (method)robotout_break,       "break",               0);
    class_addmethod(c, (method)robotout_ft1,         "ft1",     A_FLOAT,    0);
    class_addmethod(c, (method)robotout_ft2,         "ft2",     A_FLOAT,    0);
    class_addmethod(c, (method)robotout_ft3,         "ft3",     A_FLOAT,    0);
    class_addmethod(c, (method)robotout_ft4,         "ft4",     A_FLOAT,    0);
    class_addmethod(c, (method)robotout_ft5,         "ft5",     A_FLOAT,    0);
    class_addmethod(c, (method)robotout_ft6,         "ft6",     A_FLOAT,    0);
    
    
    /* you CAN'T call this from the patcher */
    class_register(CLASS_BOX, c); /* CLASS_NOBOX */
    robotout_class = c;
    
    post("Robotout object created (Writer).");
}

void robotout_assist(t_robotout *x, void *b, long m, long a, char *s)
{
    if (m == ASSIST_INLET) { // inlet
        
        switch(a) {
            case 0: sprintf(s, "Messages"); break;
            case 1: sprintf(s, "Linear X"); break;
            case 2: sprintf(s, "Linear Y"); break;
            case 3: sprintf(s, "Linear Z"); break;
            case 4: sprintf(s, "Angular X"); break;
            case 5: sprintf(s, "Angular Y"); break;
            case 6: sprintf(s, "Angular Z"); break;
        }
    }
}

void robotout_free(t_robotout *x)
{
}

void *robotout_new(t_symbol *s, long argc, t_atom *argv)
{
    t_robotout *x = (t_robotout *)object_alloc(robotout_class);
    // verify memory allocated successfully
    if (x == NULL ) {
        post( "Max is out of memory, could not create new robotout object.");
        return NULL;
    }
    
    // ip address attribute
    if(argv->a_type == A_SYM) {
        x->ip_addr = atom_getsym(argv);
    }
    
    else
        x->ip_addr = NULL;
    
    x->is_ready = false;
    
    // create pin integer inlets
    for (int i = 6; i > 0; --i) {
        // create float inlets from right to left (highest-right to 1-left):
        floatin(x, i);
        x->inlets[i-1] = 0;
    }
    
    return (x);
}

void robotout_bang(t_robotout *x) {
    
    // if connection to robot not established
    // (publisher did not advertise topic yet)
    // then establish it now
    if( x->is_ready == false) {
        if (robotout_open(x) != 0)
            return;
    }
    
    // send (publish) the new values
    ros_publish_twist(x->inlets);
    
 //   object_post((t_object *)x, "published to robot %s the message: [%f, %f, %f],[%f, %f, %f]", x->ip_addr->s_name, x->inlets[0], x->inlets[1], x->inlets[2], x->inlets[3], x->inlets[4], x->inlets[5]);
}

// Send break (all 0 velocities) to robot
void robotout_break(t_robotout *x) {
    
    // if connection to robot not established
    // (publisher did not advertise topic yet)
    // then return
    if( x->is_ready == false)
            return;
    
    for(int i=0; i<6; ++i)
        x->inlets[i] = 0;
    
    // send (publish) the new values
    ros_publish_twist(x->inlets);
    
    object_post((t_object *)x, "break");
}

int robotout_open(t_robotout *x) {
    
    if(x->ip_addr == NULL) {
        object_error((t_object *)x, "Please enter the robot's IP address");
        return -1;
    }
    
    if(x->is_ready) {
        object_post((t_object *)x, "already openned a connection to robot %s", x->ip_addr->s_name);
        return 0;
    }
    
    if( ros_init(x->ip_addr->s_name) != 0) {
        object_error((t_object *)x, "Error establishing a connection to the robot. Check your IP address.");
        return -1;
    }
    
    if( ros_advertise_twist() != 0) {
        object_error((t_object *)x, "Error communicating with the robot.");
        return -1;
    }
    
    object_post((t_object *)x, "ready to publish to robot on %s", x->ip_addr->s_name);
    post("Publishing to /cmd_vel");
    
    x->is_ready = true;
    return 0;
    
}

int robotout_close(t_robotout *x) {
    x->is_ready = false;
    return 0;
}

void robotout_ft1(t_robotout *x, Float64 f) { x->inlets[0] = f; }
void robotout_ft2(t_robotout *x, Float64 f) { x->inlets[1] = f; }
void robotout_ft3(t_robotout *x, Float64 f) { x->inlets[2] = f; }
void robotout_ft4(t_robotout *x, Float64 f) { x->inlets[3] = f; }
void robotout_ft5(t_robotout *x, Float64 f) { x->inlets[4] = f; }
void robotout_ft6(t_robotout *x, Float64 f) { x->inlets[5] = f; }


