// © Copyright The University of New South Wales 2018
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see http://www.gnu.org/licenses/.

#include "wrapper.h"
#include <arpa/inet.h>
#include "ros_lib/ros.h"
#include "ros_lib/ros/subscriber.h"
#include "ros_lib/std_msgs/String.h"
#include "ros_lib/geometry_msgs/Twist.h"
#include "ros_lib/nav_msgs/Odometry.h"
#include "ros_lib/sensor_msgs/LaserScan.h"
#include "ros_lib/sensor_msgs/Imu.h"
#include "ros_lib/turtlebot3_msgs/SensorState.h"
#include "ros_lib/sensor_msgs/JointState.h"
#include "ros_lib/tf/tfMessage.h"

extern "C" {
    
    // ROS Message
    std_msgs::String str_msg;
    geometry_msgs::Twist twist_msg;
    
    // GUI callback function
    void (*twist_callback)(void*, const void*, int i);
    void (*odometry_callback)(void*, const void*, int i);
    void (*imu_callback)(void*, const void*, int i);
    void (*laser_callback)(void*, const void*, int i);
    void (*joints_callback)(void*, const void*, int i);
    void (*sensors_callback)(void*, const void*, int i);
    void (*tf_callback)(void*, const void*, int i);
    
    void *gui_obj;
    int twist_order;
    int odometry_order;
    int laser_order;
    int imu_order;
    int sensors_order;
    int joints_order;
    int tf_order;
    
    // Ros Node
    ros::NodeHandle  nh;
    
    // ROS Publishers
    ros::Publisher pub_twist("/cmd_vel", &twist_msg);
    ros::Publisher pub_chatter("chatter", &str_msg);
    
    // ROS Subscribers
    // turn (geometry_msgs::Twist) into (**void) - c_wrapper
    void twistCb(const geometry_msgs::Twist& received_msg) {
        (*twist_callback)(gui_obj, &received_msg, twist_order);
    }
    void odometryCb(const nav_msgs::Odometry& received_msg) {
        (*odometry_callback)(gui_obj, &received_msg, odometry_order);
    }
    void laserCb(const sensor_msgs::LaserScan& received_msg) {
        (*laser_callback)(gui_obj, &received_msg, laser_order);
    }
    void imuCb(const sensor_msgs::Imu& received_msg) {
        (*imu_callback)(gui_obj, &received_msg, imu_order);
    }
    void sensorsCb(const turtlebot3_msgs::SensorState& received_msg) {
        (*sensors_callback)(gui_obj, &received_msg, sensors_order);
    }
    void jointsCb(const sensor_msgs::JointState& received_msg) {
        (*joints_callback)(gui_obj, &received_msg, joints_order);
    }
    void tfCb(const tf::tfMessage& received_msg) {
        (*tf_callback)(gui_obj, &received_msg, tf_order);
    }
    ros::Subscriber<geometry_msgs::Twist> subTwist("/cmd_vel_rc100", twistCb);
    ros::Subscriber<nav_msgs::Odometry> subOdometry("/odom", odometryCb);
    ros::Subscriber<sensor_msgs::LaserScan> subLaser("/scan", laserCb);    
    ros::Subscriber<sensor_msgs::Imu> subImu("/imu", imuCb);
    ros::Subscriber<turtlebot3_msgs::SensorState> subSensors("/sensor_state", sensorsCb);
    ros::Subscriber<sensor_msgs::JointState> subJoints("/joint_states", jointsCb);
    ros::Subscriber<tf::tfMessage> subTf("/tf", tfCb);
    
    double *twistMsg_floats(const void *msg_) {
        const geometry_msgs::Twist *msg = (geometry_msgs::Twist*)msg_; 
        double *data = (double *)malloc(6*sizeof(double));
        data[0] = msg->linear.x;
        data[1] = msg->linear.y;
        data[2] = msg->linear.z;
        data[3] = msg->angular.x;
        data[4] = msg->angular.y;
        data[5] = msg->angular.z;
        return data;
    }
    
    unsigned *odometryMsg_header_ints(const void *msg_) {
        const nav_msgs::Odometry *msg = (nav_msgs::Odometry*)msg_;
        unsigned *data = (unsigned *)malloc(3*sizeof(unsigned));
        data[0] = msg->header.seq;
        data[1] = msg->header.stamp.sec;
        data[2] = msg->header.stamp.nsec;
        return data;
    }
    
    char *odometryMsg_header_frame(const void *msg_) {
        const nav_msgs::Odometry *msg = (nav_msgs::Odometry*)msg_;
        char *data = (char *)malloc(strlen(msg->header.frame_id));
        sprintf(data, "%s", msg->header.frame_id);
        return data;
    }
    
    char *odometryMsg_childframe(const void *msg_) {
        const nav_msgs::Odometry *msg = (nav_msgs::Odometry*)msg_;
        char *data = (char *)malloc(strlen(msg->child_frame_id));
        sprintf(data, "%s", msg->child_frame_id);
        return data;
    }
    
    double *odometryMsg_floats(const void *msg_) {
        const nav_msgs::Odometry *msg = (nav_msgs::Odometry*)msg_;
        double *data = (double *)malloc(85*sizeof(double));
        int i=0;
        data[i++] = msg->pose.pose.position.x;
        data[i++] = msg->pose.pose.position.y;
        data[i++] = msg->pose.pose.position.z;
        data[i++] = msg->pose.pose.orientation.x;
        data[i++] = msg->pose.pose.orientation.y;
        data[i++] = msg->pose.pose.orientation.z;
        data[i++] = msg->pose.pose.orientation.w;
        for(int j=0; j<36; ++j)
            data[i++] = msg->pose.covariance[j];
        data[i++] = msg->twist.twist.linear.x;
        data[i++] = msg->twist.twist.linear.y;
        data[i++] = msg->twist.twist.linear.z;
        data[i++] = msg->twist.twist.angular.x;
        data[i++] = msg->twist.twist.angular.y;
        data[i++] = msg->twist.twist.angular.z;
        for(int j=0; j<36; ++j)
            data[i++] = msg->twist.covariance[j];
        return data;
    }
    
    unsigned *imuMsg_header_ints(const void *msg_) {
        const sensor_msgs::Imu *msg = (sensor_msgs::Imu*)msg_;
        unsigned *data = (unsigned *)malloc(3*sizeof(unsigned));
        data[0] = msg->header.seq;
        data[1] = msg->header.stamp.sec;
        data[2] = msg->header.stamp.nsec;
        return data;
    }

    char *imuMsg_header_frame(const void *msg_) {
        const sensor_msgs::Imu *msg = (sensor_msgs::Imu*)msg_;
        char *data = (char *)malloc(strlen(msg->header.frame_id));
        sprintf(data, "%s", msg->header.frame_id);
        return data;
    }

    double *imuMsg_floats(const void *msg_) {
        const sensor_msgs::Imu *msg = (sensor_msgs::Imu*)msg_;
        double *data = (double *)malloc(37*sizeof(double));
        int i=0;
        data[i++] = msg->orientation.x;
        data[i++] = msg->orientation.y;
        data[i++] = msg->orientation.z;
        data[i++] = msg->orientation.w;
        for(int j=0; j<9; ++j)
            data[i++] = msg->orientation_covariance[j];
        data[i++] = msg->angular_velocity.x;
        data[i++] = msg->angular_velocity.y;
        data[i++] = msg->angular_velocity.z;
        for(int j=0; j<9; ++j)
            data[i++] = msg->angular_velocity_covariance[j];
        data[i++] = msg->linear_acceleration.x;
        data[i++] = msg->linear_acceleration.y;
        data[i++] = msg->linear_acceleration.z;
        for(int j=0; j<9; ++j)
            data[i++] = msg->linear_acceleration_covariance[j];
        return data;
    }

    char *laserMsg_header_frame(const void *msg_) {
        const sensor_msgs::LaserScan *msg = (sensor_msgs::LaserScan*)msg_;
        char *data = (char *)malloc(strlen(msg->header.frame_id));
        sprintf(data, "%s", msg->header.frame_id);
        return data;
    }

    double *laserMsg_floats(const void *msg_) {
        const sensor_msgs::LaserScan *msg = (sensor_msgs::LaserScan*)msg_;
        int range_length = msg->ranges_length;
        int intensities_length = msg->intensities_length;
        int num_total = 7 + range_length + intensities_length;
        double *data = (double *)malloc(num_total*sizeof(double));
        data[0] = msg->angle_min;
        data[1] = msg->angle_max;
        data[2] = msg->angle_increment;
        data[3] = msg->time_increment;
        data[4] = msg->scan_time;
        data[5] = msg->range_min ;
        data[6] = msg->range_max ;
        int i = 7;
        data[i] = NULL;
        for(int j=0; j<range_length; ++j)
            data[i++] = msg->ranges[j];
        data[i] = NULL;
        for(int j=0; j<intensities_length; ++j)
            data[i++] = msg->intensities[j];
        return data;
    }
    
    unsigned *laserMsg_header_ints(const void *msg_) {
        const sensor_msgs::LaserScan *msg = (sensor_msgs::LaserScan*)msg_;
        unsigned *data = (unsigned *)malloc(3*sizeof(unsigned));
        data[0] = msg->header.seq;
        data[1] = msg->header.stamp.sec;
        data[2] = msg->header.stamp.nsec;
        return data;
    }
    
    int laser_ranges_length(const void *msg_) {
        const sensor_msgs::LaserScan *msg = (sensor_msgs::LaserScan*)msg_;
        return msg->ranges_length;
    }
    int laser_intensities_length(const void *msg_) {
        const sensor_msgs::LaserScan *msg = (sensor_msgs::LaserScan*)msg_;
        return  msg->intensities_length;
    }
    
    unsigned *jointsMsg_header_ints(const void *msg_) {
        const sensor_msgs::JointState *msg = (sensor_msgs::JointState*)msg_;
        unsigned *data = (unsigned *)malloc(3*sizeof(unsigned));
        data[0] = msg->header.seq;
        data[1] = msg->header.stamp.sec;
        data[2] = msg->header.stamp.nsec;
        return data;
    }

    char *jointsMsg_header_frame(const void *msg_) {
        const sensor_msgs::JointState *msg = (sensor_msgs::JointState*)msg_;
        char *data = (char *)malloc(strlen(msg->header.frame_id));
        sprintf(data, "%s", msg->header.frame_id);
        return data;
    }
    
    int jointsMsg_length(const void *msg_) {
        const sensor_msgs::JointState *msg = (sensor_msgs::JointState*)msg_;
        return msg->name_length;
    }
    
    char **jointsMsg_names(const void *msg_) {
        const sensor_msgs::JointState *msg = (sensor_msgs::JointState*)msg_;
        int length = msg->name_length;
        
        char **data = (char**)malloc(length*sizeof(char*));
        for(int i=0; i<3; ++i) {
            data[i] = (char*)malloc(sizeof(msg->name[i]));
            sprintf(data[i], "%s", msg->name[i]);
        }
        return data;
    }
    
    double *jointsMsg_floats(const void *msg_) {
        const sensor_msgs::JointState *msg = (sensor_msgs::JointState*)msg_;
        int length = msg->name_length;
        int num_total = 3 * length;
        double *data = (double *)malloc(num_total*sizeof(double));
        for(int i=0; i<length; ++i)
            data[i] = msg->position[i];
        int count = length;
        for(int i=0; i<length; ++i)
            data[i+count] = msg->velocity[i];
        count = 2*length;
        for(int i=0; i<length; ++i)
            data[i+count] = msg->effort[i];
        return data;
    }
    
    unsigned *sensorsMsg_ints(const void *msg_) {
        const turtlebot3_msgs::SensorState *msg = (turtlebot3_msgs::SensorState*)msg_;
        unsigned *data = (unsigned *)malloc(7*sizeof(unsigned));
        data[0] = msg->stamp.sec;
        data[1] = msg->stamp.nsec;
        data[2] = msg->bumper;
        data[3] = msg->cliff;
        data[4] = msg->button;
        data[5] = msg->left_encoder;
        data[6] = msg->right_encoder;
        return data;
  
        /* ROSSERIAL MESSAGE IMPLEMENTATION:
        ros::Time stamp; // sec, nsec
        uint8_t bumper;
        uint8_t cliff;
        uint8_t button;
        uint8_t left_encoder;
        uint8_t right_encoder;
        float battery;
        */
    }

    double sensorsMsg_float(const void *msg_) {
        const turtlebot3_msgs::SensorState *msg = (turtlebot3_msgs::SensorState*)msg_;
        return msg->battery;
    }
    
    unsigned *tfMsg_header_ints(const void *msg_) {
        const tf::tfMessage *msg = (tf::tfMessage*)msg_;
        unsigned *data = (unsigned *)malloc(3*sizeof(unsigned));
        data[0] = msg->transforms[0].header.seq;
        data[1] = msg->transforms[0].header.stamp.sec;
        data[2] = msg->transforms[0].header.stamp.nsec;
        return data;
    }
    
    char *tfMsg_header_frame(const void *msg_) {
        const tf::tfMessage *msg = (tf::tfMessage*)msg_;
        char *data = (char *)malloc(strlen(msg->transforms[0].header.frame_id));
        sprintf(data, "%s", msg->transforms[0].header.frame_id);
        return data;
    }
    
    char *tfMsg_childframe(const void *msg_) {
        const tf::tfMessage *msg = (tf::tfMessage*)msg_;
        char *data = (char *)malloc(strlen(msg->transforms[0].child_frame_id));
        sprintf(data, "%s", msg->transforms[0].child_frame_id);
        return data;
    }
    
    double *tfMsg_floats(const void *msg_) {
        const tf::tfMessage *msg = (tf::tfMessage*)msg_;
        double *data = (double *)malloc(7*sizeof(double));
        int i=0;
        data[i++] = msg->transforms[0].transform.translation.x;
        data[i++] = msg->transforms[0].transform.translation.y;
        data[i++] = msg->transforms[0].transform.translation.z;
        data[i++] = msg->transforms[0].transform.rotation.x;
        data[i++] = msg->transforms[0].transform.rotation.y;
        data[i++] = msg->transforms[0].transform.rotation.z;
        data[i++] = msg->transforms[0].transform.rotation.w;
        return data;
    }
    
    // ROS Communication
    int ros_init(char *rosSrvrIp) {
        if(isValidIpAddress(rosSrvrIp) == false) {
            return -1;
        }
        //nh.initNode();
        nh.initNode(rosSrvrIp);
        return 0;
    }
    
    void ros_advertise_chatter() {
        nh.advertise(pub_chatter);
        sleep(1);
    }

    int ros_advertise_twist() {
        nh.advertise(pub_twist);
        sleep(1);
        return 0;
    }

    void ros_publish_chatter(const void *msg) {
        
        if(msg == NULL) return;
        
        str_msg.data = (char*)msg;
        pub_chatter.publish( &str_msg );
        nh.spinOnce();
    }

    int ros_publish_twist(const double out_msg[6]) {
        
        twist_msg.linear.x = out_msg[0];
        twist_msg.linear.y = out_msg[1];
        twist_msg.linear.z = out_msg[2];
        
        twist_msg.angular.x = out_msg[3];
        twist_msg.angular.y = out_msg[4];
        twist_msg.angular.z = out_msg[5];
        
        pub_twist.publish( &twist_msg );
        nh.spinOnce();
        return 0;
    }

    
    void ros_subscribe_chatter(void *obj, void (*f)(void*, void**) ) {
    }
    
    void ros_subscribe_twist(void *obj, void (*f)(void*, const void*, int), int index) {
        twist_callback = f;
        gui_obj = obj;
        twist_order = index;
        nh.subscribe(subTwist);
    }
    
    void ros_subscribe_odometry(void *obj, void (*f)(void*, const void*, int), int index) {
        odometry_callback = f;
        gui_obj = obj;
        odometry_order = index;
        nh.subscribe(subOdometry);
    }

    void ros_subscribe_laser(void *obj, void (*f)(void*, const void*, int), int index) {
        laser_callback = f;
        gui_obj = obj;
        laser_order = index;
        nh.subscribe(subLaser);
    }

    void ros_subscribe_imu(void *obj, void (*f)(void*, const void*, int), int index) {
        imu_callback = f;
        gui_obj = obj;
        imu_order = index;
        nh.subscribe(subImu);
    }

    void ros_subscribe_joints(void *obj, void (*f)(void*, const void*, int), int index) {
        joints_callback = f;
        gui_obj = obj;
        joints_order = index;
        nh.subscribe(subJoints);
    }

    void ros_subscribe_sensors(void *obj, void (*f)(void*, const void*, int), int index) {
        sensors_callback = f;
        gui_obj = obj;
        sensors_order = index;
        nh.subscribe(subSensors);
    }
    
    void ros_subscribe_tf(void *obj, void (*f)(void*, const void*, int), int index) {
        tf_callback = f;
        gui_obj = obj;
        tf_order = index;
        nh.subscribe(subTf);
    }
    
    void ros_listen(void *data) {
        while(1) {
            nh.spinOnce();
            sleep(1);
        }
    }

    int isValidIpAddress(char *ipAddress) {
        struct sockaddr_in sa;
        return inet_pton(AF_INET, ipAddress, &(sa.sin_addr));
    }
    
    
}

/*
    double twistMsg_get_linera_x(const void *msg) {
        const geometry_msgs::Twist *t_msg = (const geometry_msgs::Twist*)msg;
        return t_msg->linear.x;
    }
    
    double twistMsg_get_linera_y(const void *msg) {
        const geometry_msgs::Twist *t_msg = (const geometry_msgs::Twist*)msg;
        return t_msg->linear.x;
    }
    double twistMsg_get_linera_z(const void *msg) {
        const geometry_msgs::Twist *t_msg = (const geometry_msgs::Twist*)msg;
        return t_msg->linear.x;
    }
    
    double twistMsg_get_angular_x(const void *msg) {
        const geometry_msgs::Twist *t_msg = (const geometry_msgs::Twist*)msg;
        return t_msg->linear.x;
    }
    
    double twistMsg_get_angular_y(const void *msg) {
        const geometry_msgs::Twist *t_msg = (const geometry_msgs::Twist*)msg;
        return t_msg->linear.x;
    }
    
    double twistMsg_get_angular_z(const void *msg) {
        const geometry_msgs::Twist *t_msg = (const geometry_msgs::Twist*)msg;
        return t_msg->linear.x;
    }
    
}

const double* twistMsgToC(const geometry_msgs::Twist& received_msg)  {
    double *msg;
    msg = (double*)malloc(6*sizeof(double));
    msg[0] = received_msg.linear.x;
    msg[1] = received_msg.linear.y;
    msg[2] = received_msg.linear.z;
    msg[3] = received_msg.angular.x;
    msg[4] = received_msg.angular.y;
    msg[5] = received_msg.angular.z;
    return msg;
}


const double** twistMsgToC(const geometry_msgs::Twist& received_msg)  {
    const double **msg;
    msg = (const double**)malloc(6*sizeof(const double*));
    msg[0] = &(received_msg.linear.x);
    msg[1] = &(received_msg.linear.y);
    msg[2] = &(received_msg.linear.z);
    msg[3] = &(received_msg.angular.x);
    msg[4] = &(received_msg.angular.y);
    msg[5] = &(received_msg.angular.z);
    return msg;//(const void**)msg;
}


 ros::Subscriber_ *sub;
 void msgCb(const geometry_msgs::Twist& received_msg) {
 
 // turn (geometry_msgs::Twist) into (**void) - c_wrapper
 // turn (**void) into (t_atom*) and output
 const void** twist_msg = (const void**)twistMsgToC(received_msg);
 (*gui_callback)(gui_obj, twist_msg, out_order);
 }
 
 // turn (geometry_msgs::Twist) into (**void) - c_wrapper
 void twistCb1(const geometry_msgs::Twist& received_msg) {
 const void** twist_msg = (const void**)twistMsgToC(received_msg);
 (*gui_callback)(gui_obj, twist_msg, out_order);
 }
 
*/
